/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package publisher

import (
	"strings"
	"time"

	"google.golang.org/protobuf/proto"
	"google.golang.org/protobuf/reflect/protoreflect"

	"carbonblack.com/protocols/psc/ems/assetupdater"
	producer "github.com/a8m/kinesis-producer"
	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/credentials"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/kinesis"
	"github.com/aws/aws-sdk-go/service/kinesis/kinesisiface"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/wavefronthq/go-metrics-wavefront/reporting"

	"gitlab.bit9.local/psc/platform/orgkey-go/v3"

	"github.com/aws/aws-sdk-go/aws/ec2metadata"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

func NewKplMetrics(metricsFactory *stats.PrometheusMetricsFactory) *KplMetrics {
	kplPublishedCount := metricsFactory.NewCounterVec("published_count", "kpl", []string{"asset_env"})
	kplFailedCount := metricsFactory.NewCounterVec("failed_count", "kpl", []string{"asset_env", "asset_urn"})
	return &KplMetrics{KplPlubishedCount: kplPublishedCount, KplFailureCount: kplFailedCount}
}

type KplMetrics struct {
	KplPlubishedCount *prometheus.CounterVec
	KplFailureCount   *prometheus.CounterVec
}

// KinesisPublisher kinesis publisher wrapper for aws-sdk publisher and kinesis client
type KinesisPublisher struct {
	producer        *producer.Producer
	config          *config.Configurations
	log             *zerolog.Logger
	client          kinesisiface.KinesisAPI
	metricsReporter reporting.WavefrontMetricsReporter
	*KplMetrics
}

// IKinesisPublisher describes methods for IKinesisPublisher
type IKinesisPublisher interface {
	StartProducer()
	StopProducer()
	PutRecords(deviceLastContactTimeRecords []*assetupdater.AssetDataChangeEvent) error
	StreamExists(streamName *string) bool
}

// NewKinesisPublisher -- create a new producer
func NewKinesisPublisher(logger *zerolog.Logger, cfg *config.Configurations, wfReporter reporting.WavefrontMetricsReporter, metricsFactory *stats.PrometheusMetricsFactory) *KinesisPublisher {
	return NewKinesisPublisherWithMetrics(logger, cfg, wfReporter, NewKplMetrics(metricsFactory))
}

// NewKinesisPublisherWithMetrics -- create a new producer w/ explicit metrics struct
func NewKinesisPublisherWithMetrics(logger *zerolog.Logger, cfg *config.Configurations, wfReporter reporting.WavefrontMetricsReporter, metrics *KplMetrics) *KinesisPublisher {
	kinesisClient := createKinesisClient(cfg, logger)
	return NewKinesisPublisherWithClient(kinesisClient, logger, cfg, wfReporter, metrics)
}

// NewKinesisPublisherWithClient -- create a new producer with a kinesis client
func NewKinesisPublisherWithClient(kinesisClient kinesisiface.KinesisAPI, logger *zerolog.Logger, cfg *config.Configurations, wfReporter reporting.WavefrontMetricsReporter, kplMetrics *KplMetrics) *KinesisPublisher {
	kinesisProducer := producer.New(&producer.Config{
		StreamName:         cfg.AssetRegistryStreamName,
		Client:             kinesisClient,
		AggregateBatchSize: cfg.AggregationBatchSize,
		BatchCount:         cfg.ProducerBatchCount,
		FlushInterval:      time.Duration(cfg.FlushIntervalSec) * time.Second,
	})

	// Handle failures
	go func() {
		for r := range kinesisProducer.NotifyFailures() {
			// r contains `Data`, `PartitionKey` and `Error()`
			logger.Error().Msgf("Error while posting to kinesis:- %v", r)

			// Update WF metrics on any Failure
			event := ProtoUnmarshal(r.Data, nil)
			tags := make(map[string]string)
			if event != nil {
				tags["env"] = GetEnvFromAssetUrn(event.AssetUrn)
				tags["asset_urn"] = event.AssetUrn
				tags["msc"] = stats.Mosaic2Tag
			}
			stats.UpdateCounter(wfReporter, "kpl.asset-registry.failure", tags, 1)
			kplMetrics.KplFailureCount.WithLabelValues(tags["env"], tags["asset_urn"]).Inc()
		}
	}()

	return &KinesisPublisher{
		producer:        kinesisProducer,
		config:          cfg,
		log:             logger,
		client:          kinesisClient,
		metricsReporter: wfReporter,
		KplMetrics:      kplMetrics,
	}
}

// GetEnvFromAssetUrn - Extracts Env name from Asset Urn
func GetEnvFromAssetUrn(urn string) string {
	splitUrn := strings.Split(urn, ":")
	if len(splitUrn) >= 3 {
		urnOrgKey := splitUrn[2]
		orgKey, err := orgkey.Decode(urnOrgKey)
		if err != nil {
			return orgKey.Env()
		}
		return "UNKNOWN"
	}
	return "UNKNOWN"
}

// StartProducer - starts the producer daemon
func (lct *KinesisPublisher) StartProducer() {
	lct.producer.Start()
	lct.log.Info().Msgf("Asset Registry Kinesis publisher Started")
}

// StopProducer - stops the producer daemon
func (lct *KinesisPublisher) StopProducer() {
	lct.producer.Stop()
	lct.log.Info().Msgf("Asset Registry Kinesis publisher Stopped")
}

// StreamExists - checks if Stream exists or not
func (lct *KinesisPublisher) StreamExists(streamName *string) bool {
	describeStreamReq := &kinesis.DescribeStreamInput{
		ExclusiveStartShardId: nil,
		Limit:                 nil,
		StreamName:            streamName,
	}

	dso, err := lct.client.DescribeStream(describeStreamReq)
	if err != nil {
		lct.log.Error().Msgf("Asset Registry Stream (%s) doesn't exist, err = %v ", *streamName, err)
		return false
	}
	lct.log.Debug().Msgf("Asset Registry Kinesis Stream Info - %v", dso.StreamDescription)
	return true
}

// PutRecords - used for putting AssetDataChangeEvent records
func (lct *KinesisPublisher) PutRecords(assetDataEventRecord []*assetupdater.AssetDataChangeEvent, eventCountMap map[string]int64) {
	tags := map[string]string{"env": lct.config.CbdEnvName, "msc": stats.Mosaic2Tag}
	for _, record := range assetDataEventRecord {
		dataBytes := ProtoMarshal(record, lct.log)
		err := lct.producer.Put(dataBytes, record.AssetUrn)
		if err != nil {
			lct.log.Error().Msgf("Error while posting lastContactTime batch to Asset Registry stream: %v, asset_urn: %v", err, record.AssetUrn)
			stats.UpdateCounter(lct.metricsReporter, "kpl.asset-registry.failed", tags, 1)
			lct.KplFailureCount.WithLabelValues(lct.config.CbdEnvName, "").Inc()
		} else {
			lct.log.Info().Msgf("Pushing %+v to asset registry stream ...", record)
		}
	}

	for _, count := range eventCountMap {
		stats.UpdateCounter(lct.metricsReporter, "kpl.asset-registry.published", tags, count)
		lct.KplPlubishedCount.WithLabelValues(lct.config.CbdEnvName).Add(float64(count))
	}
}

// ProtoMarshal - using proto marshal to convert to byte[]
func ProtoMarshal(event protoreflect.ProtoMessage, logger *zerolog.Logger) []byte {
	if logger == nil {
		logger = &log.Logger
	}

	marshalledBytes, err := proto.Marshal(event)
	if err != nil {
		logger.Error().Msgf("Proto Marshal failed, error - %v", err)
	}
	return marshalledBytes
}

// ProtoUnmarshal - Method used to decode AssetUpdater byte data from Kinesis stream, used for test only
func ProtoUnmarshal(p []byte, logger *zerolog.Logger) *assetupdater.AssetDataChangeEvent {
	if logger == nil {
		logger = &log.Logger
	}

	assetDataChangeEvent := &assetupdater.AssetDataChangeEvent{}
	err := proto.Unmarshal(p, assetDataChangeEvent)
	if err != nil {
		logger.Error().Msgf("Proto Unmarshall failed, error - %v", err)

		err := proto.Unmarshal(p, assetDataChangeEvent)
		if err != nil {
			logger.Error().Msgf("Error while encoding AssetRegistry lastContactTime event data - %v", err)
		}
	}
	logger.Printf("Uncompressed size (bytes): %v", len(p))
	return assetDataChangeEvent
}

const azPlaceHolder = "<AZ>"

// createKinesisClient - Creates Kinesis Client for the given region in config
func createKinesisClient(conf *config.Configurations, logger *zerolog.Logger) *kinesis.Kinesis {
	awsConfig := aws.NewConfig().WithRegion(conf.RegionName).WithCredentialsChainVerboseErrors(true)

	if conf.IsLocalEnv() {
		staticCreds := credentials.NewStaticCredentials("test", "test", "test")
		awsConfig = awsConfig.WithEndpoint(conf.AwsEndpoint).WithCredentials(staticCreds)
	}

	sess, err := session.NewSession(awsConfig)
	if err != nil {
		logger.Fatal().Msgf("Not able to create session with kinesis - %v", err)
	}

	if conf.IsLocalEnv() || !strings.Contains(conf.AssetRegistryKinesisEndpoint, azPlaceHolder) {
		return kinesis.New(sess)
	}

	metaData, err := ec2metadata.New(sess).GetInstanceIdentityDocument()
	if err != nil {
		logger.Fatal().Msgf("UNABLE TO USE METADATA API TO DETECT AZ %v", err)
	}
	az := metaData.AvailabilityZone

	kinesisEndpoint := strings.Replace(conf.AssetRegistryKinesisEndpoint, azPlaceHolder, az, 1)

	return kinesis.New(sess, &aws.Config{Endpoint: &kinesisEndpoint})
}
