package stats

import (
	"net/http"
	"os"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/collectors"
	"github.com/prometheus/client_golang/prometheus/promauto"
	"github.com/prometheus/client_golang/prometheus/promhttp"
)

const (
	metricsPrefix = "ems_last_contact_updater_"
	envLabel      = "env"
	sourceLabel   = "source"
)

type PrometheusMetricsFactory struct {
	gatherer   prometheus.Gatherer
	registerer prometheus.Registerer
	factory    *promauto.Factory
}

func NewPrometheusMetrticsFactory(envName string) *PrometheusMetricsFactory {
	return newPrefixWrappedRegistry(metricsPrefix, envName)
}

func newPrefixWrappedRegistry(prefix, env string) *PrometheusMetricsFactory {
	registry := prometheus.NewRegistry()
	wrapped := prometheus.WrapRegistererWithPrefix(prefix, registry)
	wrapped = prometheus.WrapRegistererWith(prometheus.Labels{
		envLabel: env, sourceLabel: func() string {
			hostname, err := os.Hostname()
			if err != nil {
				hostname = "unknown"
			}

			return hostname
		}(),
	}, wrapped)
	promFactory := promauto.With(wrapped)
	return &PrometheusMetricsFactory{registerer: wrapped, gatherer: registry, factory: &promFactory}
}

func getMetricsOpts(name, subsystem string) prometheus.Opts {
	return prometheus.Opts{Name: name, Help: name, Subsystem: subsystem}
}

func (f *PrometheusMetricsFactory) NewCounterVec(name, subsystem string, tags []string) *prometheus.CounterVec {
	return f.factory.NewCounterVec(prometheus.CounterOpts(getMetricsOpts(name, subsystem)), tags)
}

func (f *PrometheusMetricsFactory) NewGaugeVec(name, subsystem string, tags []string) *prometheus.GaugeVec {
	return f.factory.NewGaugeVec(prometheus.GaugeOpts(getMetricsOpts(name, subsystem)), tags)
}

func (f *PrometheusMetricsFactory) HttpHandler() http.Handler {
	return promhttp.HandlerFor(f.gatherer, promhttp.HandlerOpts{})
}

func (f *PrometheusMetricsFactory) RegisterGoMetrics() {
	f.registerer.MustRegister(collectors.NewGoCollector(
		collectors.WithGoCollectorRuntimeMetrics(collectors.MetricsMemory),
	))
}
